from __future__ import print_function, absolute_import

from .grid import GRID
from .prid import PRID
from .ilids import iLIDS
from .viper import VIPeR
from .cuhk01 import CUHK01
from .cuhk02 import CUHK02
from .cuhk03 import CUHK03
from .msmt17 import MSMT17
from .cuhksysu import CUHKSYSU
from .sensereid import SenseReID
from .market1501 import Market1501
from .dukemtmcreid import DukeMTMCreID
from .university1652 import University1652
